@@ -8,29 +8,31 @@ from time import sleep
 from time import time
 from typing import Optional, Tuple, Any, List, Dict
 from typing import Mapping
-
 import boto3.session
 import requests
 from dateutil.parser import parse as parse_time
 import pytz
+from pytz import utc
 from botocore.utils import SSOTokenLoader
+from botocore.config import Config as BotoConfig
 from botocore.credentials import JSONFileCache
-from botocore.exceptions import ClientError
+from botocore.exceptions import ClientError, ConnectTimeoutError, ReadTimeoutError
 from raft import Context
 from raft.tasks import Task
 from ..base.utils import notice, notice_end, get_context_value, notice_level
 
-g_cache_dir_noticed = False
+g_sso_cache_dir_noticed = False
+g_cli_cache_dir_noticed = False
 
 
-def cache_dir():
+def sso_cache_dir():
     """
     gets the aws cache dir, which is user specific as ~/.aws/sso/cache
     and ensures that the directories exist with permissions 0700
     all the way down
     """
-    global g_cache_dir_noticed
-    if not g_cache_dir_noticed:
+    global g_sso_cache_dir_noticed
+    if not g_sso_cache_dir_noticed:
         notice('checking for cache dir')
     d = os.path.expanduser('~')
     dirs = [ '.aws', 'sso', 'cache' ]
@@ -41,23 +43,70 @@ def cache_dir():
     if not os.path.exists(d):
         for x in rg:
             os.makedirs(x, 0o700, exist_ok=True)
-    if not g_cache_dir_noticed:
+    if not g_sso_cache_dir_noticed:
         notice_end(d, 'green')
-        g_cache_dir_noticed = True
+        g_sso_cache_dir_noticed = True
     return d
 
 
-def client_credentials(sso_oidc, force=False):
+def cli_cache_dir():
+    """
+    gets the aws cache dir, which is user specific as ~/.aws/sso/cache
+    and ensures that the directories exist with permissions 0700
+    all the way down
+    """
+    global g_cli_cache_dir_noticed
+    if not g_cli_cache_dir_noticed:
+        notice('checking for cache dir')
+    d = os.path.expanduser('~')
+    dirs = [ '.aws', 'cli', 'cache' ]
+    rg = []
+    for x in dirs:
+        d = os.path.join(d, x)
+        rg.append(d)
+    if not os.path.exists(d):
+        for x in rg:
+            os.makedirs(x, 0o700, exist_ok=True)
+    if not g_cli_cache_dir_noticed:
+        notice_end(d, 'green')
+        g_cli_cache_dir_noticed = True
+    return d
+
+
+def timeout_config():
+    c = BotoConfig()
+    c.read_timeout = 5
+    c.connect_timeout = 5
+    c.retries = {
+        'max_attempts': 2,
+        'mode': 'standard'
+    }
+    return c
+
+
+def to_iso(t):
+    """
+    converts a timestamp to an iso format string
+    :param t: the timestamp to convert
+    :return: the iso format string
+    """
+    st = datetime.fromtimestamp(t, utc)
+    st = st.strftime('%Y-%m-%dT%H:%M:%SZ')
+    return st
+
+
+def client_credentials(sso_oidc, start_url, sso_region, force=False):
     """
     performs the first step of the 7-step oidc handshake by
     registering the client and caching client credentials
     :param sso_oidc: the boto3 sso-oidc client
+    :param str sso_region: the region name in aws -- this is used for caching
     :param bool force: specify this flag to bypass the cache
     :return: dict
     """
-    cache = SsoTokenSaver('convocations')
+    cache = SsoTokenSaver(start_url)
     message = 'client credentials'
-    expiration_key = 'clientSecretExpiresAt'
+    expiration_key = 'registrationExpiresAt'
     if not force:
         creds = cache.load(expiration_key, message) or {}
         if creds:
@@ -69,7 +118,12 @@ def client_credentials(sso_oidc, force=False):
     )
     notice_end()
     notice('caching response')
-    cache.save(creds)
+    creds = cache.save(
+        region=sso_region,
+        client_id=creds['clientId'],
+        client_secret=creds['clientSecret'],
+        registration_expires_at=to_iso(creds['clientSecretExpiresAt']),
+    )
     notice_end()
     return creds
 
@@ -82,24 +136,45 @@ class SsoTokenSaver(SSOTokenLoader):
     """
     def __init__(self, start_url):
         super().__init__()
-        self.token_cache = JSONFileCache(cache_dir())
+        self.token_cache = JSONFileCache(sso_cache_dir())
         self.start_url = start_url
         try:
-            self.cache_key = self._generate_cache_key(start_url)
-        except:
             self.cache_key = self._generate_cache_key(start_url, None)
+        except:
+            self.cache_key = self._generate_cache_key(start_url)
 
-    def save(self, response):
+    def save(self, region=None, access_token=None, expires_at=None, client_id=None, client_secret=None, registration_expires_at=None):
         """
-        caches a token response or a client registration response
-        :param dict response: the dict response returned from a call
-            to `register_client` or `create_token`
-        :return:
+        {
+            "startUrl": "...",
+            "region": "us-east-1",
+            "accessToken": "...",
+            "expiresAt": "...",
+            "clientId": "...",
+            "clientSecret": "...",
+            "registrationExpiresAt": "2025-09-02T06:56:06Z"
+        }
         """
         cache_key = self.cache_key
-        self.token_cache[cache_key] = response
+        try:
+            current = self.token_cache[cache_key]
+        except KeyError:
+            current = {}
+        values = {
+            'startUrl': self.start_url,
+            'region': region,
+            'accessToken': access_token,
+            'expiresAt': expires_at,
+            'clientId': client_id,
+            'clientSecret': client_secret,
+            'registrationExpiresAt': registration_expires_at,
+        }
+        values = { k: v for k, v in values.items() if v is not None }
+        current.update(values)
+        self.token_cache[cache_key] = current
+        return current
 
-    def load(self, expiration_key, message='checking expiration'):
+    def load(self, expiration_key, message='checking expiration', alt_expiration_key=None):
         """
         loads the cached response from the cache and checks its expiration.
         if the response is still valid (i.e., not expired), the response is
@@ -112,7 +187,12 @@ class SsoTokenSaver(SSOTokenLoader):
             notice(message)
             creds = self.token_cache[self.cache_key]
             expires_at = creds.get(expiration_key)
+            if not expires_at and alt_expiration_key:
+                expires_at = creds.get(alt_expiration_key)
             if expires_at:
+                if isinstance(expires_at, str):
+                    expires_at = parse_time(expires_at)
+                    expires_at = expires_at.timestamp()
                 now = int(time())
                 delta = expires_at - now
                 is_current = delta > 0
@@ -127,7 +207,7 @@ class SsoTokenSaver(SSOTokenLoader):
 
 class SsoCredentialSaver:
     def __init__(self, start_url, role_name, account_id):
-        self.token_cache = JSONFileCache(cache_dir())
+        self.token_cache = JSONFileCache(cli_cache_dir())
         self.start_url = start_url
         self.role_name = role_name
         self.account_id = account_id
@@ -149,7 +229,6 @@ class SsoCredentialSaver:
         return h.replace(os.path.sep, '_')
 
     def _parse_timestamp(self, timestamp_ms):
-        from pytz import utc
         # fromtimestamp expects seconds so: milliseconds / 1000 = seconds
         timestamp_seconds = timestamp_ms / 1000.0
         timestamp = datetime.fromtimestamp(timestamp_seconds, utc)
@@ -161,6 +240,7 @@ class SsoCredentialSaver:
         self.token_cache[cache_key] = {
             'ProviderType': 'sso',
             'Credentials': {
+                'AccountId': self.account_id,
                 'AccessKeyId': credentials['accessKeyId'],
                 'SecretAccessKey': credentials['secretAccessKey'],
                 'SessionToken': credentials['sessionToken'],
@@ -204,17 +284,26 @@ def device_authorization(sso_oidc, client_creds, start_url, force=False):
     """
     cache = SsoTokenSaver(start_url)
     message = 'cached device authorization'
-    expiration_key = 'accessTokenExpiration'
+    expiration_key = 'expiresAt'
+    alt_expiration_key = 'accessTokenExpiration'
     if not force:
-        creds = cache.load(expiration_key, message) or {}
+        creds = cache.load(expiration_key, message, alt_expiration_key) or {}
         if creds:
             return False, creds
-    notice('device authorization')
-    response = sso_oidc.start_device_authorization(
-        clientId=client_creds['clientId'],
-        clientSecret=client_creds['clientSecret'],
-        startUrl=start_url,
-    )
+    response = None
+    for x in range(5):
+        notice('device authorization')
+        try:
+            response = sso_oidc.start_device_authorization(
+                clientId=client_creds['clientId'],
+                clientSecret=client_creds['clientSecret'],
+                startUrl=start_url,
+            )
+            break
+        except Exception as ex:
+            notice_end(ex)
+            if x == 4:
+                raise
     device_code = response['deviceCode']
     url = response['verificationUriComplete']
     expires_in = response['expiresIn']
@@ -233,8 +322,11 @@ def device_authorization(sso_oidc, client_creds, start_url, force=False):
                 clientSecret=client_creds['clientSecret'],
             )
             creds['accessTokenExpiration'] = creds['expiresIn'] + time()
-            creds['expiresAt'] = creds['accessTokenExpiration']
-            cache.save(creds)
+            creds['expiresAt'] = datetime.fromtimestamp(creds['accessTokenExpiration'], utc).isoformat('T', 'seconds')
+            creds = cache.save(
+                access_token=creds['accessToken'],
+                expires_at=to_iso(creds['accessTokenExpiration']),
+            )
             notice_end()
             return True, creds
         except sso_oidc.exceptions.AuthorizationPendingException:
@@ -335,6 +427,19 @@ def unpack_sso_config(ctx: Context, sso_config: dict, role: str) -> Tuple[str, s
     return account_id, sso_region, start_url, role
 
 
+def load_boto3_sso_config(profile: str):
+    from botocore.session import Session as CoreSession
+    core_session = CoreSession(profile=profile)
+    profiles = core_session.full_config.get('profiles', {})
+    sso_config = profiles.get(profile, {})
+    sso_session = sso_config.get('sso_session') or {}
+    if sso_session:
+        sessions = core_session.full_config.get('sso_sessions', {})
+        sso_sessions_config = sessions.get(sso_session, {})
+        sso_config.update(sso_sessions_config)
+    return sso_config
+
+
 def make_sso_session(
         ctx: Optional[Context] = None,
         role: Optional[str] = None,
@@ -346,14 +451,11 @@ def make_sso_session(
     performs the 7-step oidc handshake to create an sso session with aws.
     see explanation here https://stackoverflow.com/a/71850591/6043170
     """
-    from botocore.session import Session as CoreSession
     from boto3 import Session
     if session:
         return session
     if profile and not sso_config:
-        core_session = CoreSession(profile=profile)
-        profiles = core_session.full_config.get('profiles', {})
-        sso_config = profiles.get(profile, {})
+        sso_config = load_boto3_sso_config(profile)
     sso_config = sso_config or {}
     t = unpack_sso_config(ctx, sso_config, role)
     account_id, sso_region, start_url, role = t
@@ -361,8 +463,8 @@ def make_sso_session(
     profile = profile or get_context_value(ctx, 'aws.profile')
 
     session = Session(region_name=sso_region)
-    sso_oidc = session.client('sso-oidc')
-    client_creds = client_credentials(sso_oidc, force_sso)
+    sso_oidc = session.client('sso-oidc', config=timeout_config())
+    client_creds = client_credentials(sso_oidc, start_url, sso_region, force_sso)
     reauth, device_auth = device_authorization(
         sso_oidc, client_creds, start_url, force_sso)
     if not device_auth:
@@ -400,14 +502,18 @@ def get_account_id(session: boto3.session.Session) -> Optional[str]:
     """
     account_id = getattr(session, 'account_id', None)
     if not account_id:
-        sts = session.client('sts')
+        sts = session.client('sts', config=timeout_config())
         notice('account id')
-        data = sts.get_caller_identity()
-        account_id = session.account_id = data['Account']
-        notice_end(data['Account'], 'yellow', attrs=[ 'bold' ])
-        notice('username')
-        session.whoami = data['Arn'].rsplit('/', 1)[-1]
-        notice_end(session.whoami, 'blue', attrs=[ 'underline' ])
+        try:
+            data = sts.get_caller_identity()
+            account_id = session.account_id = data['Account']
+            notice_end(data['Account'], 'yellow', attrs=[ 'bold' ])
+            notice('username')
+            session.whoami = data['Arn'].rsplit('/', 1)[-1]
+            notice_end(session.whoami, 'blue', attrs=[ 'underline' ])
+        except (ReadTimeoutError, ConnectTimeoutError):
+            notice_end(False)
+            return None
     return account_id
 
 
@@ -421,11 +527,24 @@ def get_alias(session: boto3.session.Session) -> Optional[str]:
     alias = getattr(session, 'alias', None)
     if not alias:
         notice('alias')
-        iam = session.client('iam')
-        aliases = iam.list_account_aliases()
-        aliases = aliases['AccountAliases']
-        alias = session.alias = aliases[0] if aliases else 'unknown'
-        notice_end(f'{session.alias}', 'yellow', attrs=[ 'bold' ])
+        iam = session.client('iam', config=timeout_config())
+        try:
+            aliases = iam.list_account_aliases()
+            aliases = aliases['AccountAliases']
+            alias = session.alias = aliases[0] if aliases else 'unknown'
+            notice_end(f'{alias}', 'yellow', attrs=[ 'bold' ])
+        except (ConnectTimeoutError, ReadTimeoutError):
+            notice_end(False)
+            alias = alias or 'zeus'
+        notice('account name')
+        account = session.client('account', config=timeout_config())
+        try:
+            result = account.get_account_information()
+            alias = session.alias = alias or result['AccountName']
+            notice_end(f'{alias}', 'yellow', attrs=[ 'bold' ])
+        except (ConnectTimeoutError, ReadTimeoutError, AttributeError):
+            notice_end(False)
+            alias = alias or 'zeus'
     return alias
 
 
@@ -452,6 +571,12 @@ def make_session(profile, session=None, region=None):
 
 
 class AwsTask(Task):
+    def pre_call(self, ctx, kwargs):
+        pass
+
+    def post_call(self, result):
+        return result
+
     def __call__(self, *args, **kwargs):
         brief = kwargs.get('brief')
         notice_level(brief=brief)
@@ -474,11 +599,8 @@ class AwsTask(Task):
         sso_config = None
         is_sso = get_context_value(ctx, 'aws.start_url')
         if not is_sso:
-            from botocore.session import Session as CoreSession
             if profile:
-                core_session = CoreSession(profile=profile)
-                profiles = core_session.full_config.get('profiles', {})
-                sso_config = profiles.get(profile, {})
+                sso_config = load_boto3_sso_config(profile)
                 is_sso = sso_config.get('sso_start_url')
                 is_sso = is_sso and sso_config.get('sso_account_id')
                 is_sso = is_sso and sso_config.get('sso_region')
@@ -495,7 +617,9 @@ class AwsTask(Task):
             setattr(ctx, 'session', session)
         kwargs['session'] = session
         kwargs['prefix'] = prefix
-        return super().__call__(*args, **kwargs)
+        self.pre_call(ctx, kwargs)
+        result = super().__call__(*args, **kwargs)
+        return self.post_call(result)
 
 
 def get_tag(p, key):
