@@ -478,7 +478,7 @@ def test(
     except:
         pass
     client = session.client('cloudformation')
-    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M')
+    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M%S')
     change_set_type = 'UPDATE' if stack_exists else 'CREATE'
     template_url, template_body, values = parameters(
         ctx, session, prefix, stack_name, filename, not stack_exists)
@@ -554,7 +554,7 @@ def parameterized_test(ctx, session, prefix, stack_name,
     except:
         pass
     client = session.client('cloudformation')
-    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M')
+    name = datetime.now().strftime(stack_name + '-%Y%m%d-%H%M%S')
     change_set_type = 'UPDATE' if stack_exists else 'CREATE'
     template_url, template_body, values = parameterized(
         ctx, session, prefix, stack_name, filename, not stack_exists)
