@@ -16,11 +16,29 @@ def ip_ranges(ctx, region='us-east-2', service='amazon'):
     prefixes = data['prefixes']
     notice_end(f'{len(prefixes)}')
     notice('filtering for region / service')
+    service = service.lower()
     i = 0
     rows = []
     for x in prefixes:
-        if x['region'] == region and x['service'].lower() == service.lower():
+        if x['region'] == region and x['service'].lower() == service:
             rows.append([ x['ip_prefix'] ])
             i += 1
     notice_end(f'{i}')
     print_table([ 'cidr' ], rows)
+
+
+@task
+def ip_range_services(ctx):
+    """
+    filters aws public ip ranges by region and service
+    """
+    notice('getting ip ranges')
+    url = 'https://ip-ranges.amazonaws.com/ip-ranges.json'
+    response = requests.get(url, timeout=10)
+    data = response.json()
+    services = { x['service'].lower() for x in data['prefixes'] }
+    services = list(services)
+    services.sort()
+    services = [ [ x ] for x in services ]
+    notice_end(f'{len(services)}')
+    print_table([ 'name' ], services)
